package dev.shadowsoffire.apotheosis.adventure.boss;

import dev.shadowsoffire.placebo.codec.PlaceboCodecs;
import net.minecraft.world.level.BlockAndTintGetter;
import net.minecraft.world.level.levelgen.Heightmap;
import net.minecraft.util.Mth;
import com.mojang.serialization.Codec;
import java.util.function.BiPredicate;
import it.unimi.dsi.fastutil.objects.ObjectIterator;
import net.minecraft.world.level.saveddata.SavedData;
import net.minecraft.world.level.LevelReader;
import java.util.Iterator;
import net.minecraft.nbt.CompoundTag;
import net.minecraftforge.event.server.ServerStartedEvent;
import net.minecraftforge.event.TickEvent;
import net.minecraft.server.level.ServerLevel;
import net.minecraftforge.event.entity.EntityJoinLevelEvent;
import dev.shadowsoffire.placebo.codec.CodecProvider;
import javax.annotation.Nullable;
import java.util.function.Function;
import net.minecraftforge.eventbus.api.EventPriority;
import net.minecraftforge.eventbus.api.SubscribeEvent;
import net.minecraftforge.network.simple.SimpleChannel;
import net.minecraft.network.chat.TextColor;
import net.minecraft.server.network.ServerGamePacketListenerImpl;
import net.minecraft.world.entity.Mob;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.level.ServerLevelAccessor;
import net.minecraft.util.RandomSource;
import dev.shadowsoffire.placebo.network.PacketDistro;
import dev.shadowsoffire.apotheosis.adventure.client.BossSpawnMessage;
import dev.shadowsoffire.apotheosis.Apotheosis;
import net.minecraft.network.protocol.Packet;
import net.minecraft.network.chat.Component;
import net.minecraft.network.protocol.game.ClientboundSetActionBarTextPacket;
import net.minecraft.server.level.ServerPlayer;
import net.minecraft.world.phys.Vec3;
import net.minecraft.world.entity.EntityType;
import net.minecraft.world.level.LevelAccessor;
import net.minecraft.world.entity.Entity;
import net.minecraft.world.entity.LivingEntity;
import dev.shadowsoffire.apotheosis.adventure.AdventureModule;
import dev.shadowsoffire.apotheosis.adventure.compat.GameStagesCompat;
import net.minecraft.world.level.Level;
import dev.shadowsoffire.placebo.reload.WeightedDynamicRegistry;
import java.util.function.Predicate;
import net.minecraft.core.BlockPos;
import dev.shadowsoffire.apotheosis.adventure.AdventureConfig;
import org.apache.commons.lang3.tuple.Pair;
import net.minecraftforge.eventbus.api.Event;
import net.minecraft.world.entity.monster.Monster;
import net.minecraft.world.entity.MobSpawnType;
import net.minecraftforge.event.entity.living.MobSpawnEvent;
import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap;
import net.minecraft.resources.ResourceLocation;
import it.unimi.dsi.fastutil.objects.Object2IntMap;

public class BossEvents
{
    public Object2IntMap<ResourceLocation> bossCooldowns;
    
    public BossEvents() {
        this.bossCooldowns = (Object2IntMap<ResourceLocation>)new Object2IntOpenHashMap();
    }
    
    @SubscribeEvent(priority = EventPriority.LOW)
    public void naturalBosses(final MobSpawnEvent.FinalizeSpawn e) {
        if (e.getSpawnType() == MobSpawnType.NATURAL || e.getSpawnType() == MobSpawnType.CHUNK_GENERATION) {
            final LivingEntity entity = (LivingEntity)e.getEntity();
            final RandomSource rand = e.getLevel().m_213780_();
            if (this.bossCooldowns.getInt((Object)entity.m_9236_().m_46472_().m_135782_()) <= 0 && !e.getLevel().m_5776_() && entity instanceof Monster && e.getResult() != Event.Result.DENY) {
                final ServerLevelAccessor sLevel = e.getLevel();
                final ResourceLocation dimId = sLevel.m_6018_().m_46472_().m_135782_();
                final Pair<Float, BossSpawnRules> rules = AdventureConfig.BOSS_SPAWN_RULES.get(dimId);
                if (rules == null) {
                    return;
                }
                if (rand.m_188501_() <= rules.getLeft() && rules.getRight().test(sLevel, BlockPos.m_274561_(e.getX(), e.getY(), e.getZ()))) {
                    final Player player = sLevel.m_45924_(e.getX(), e.getY(), e.getZ(), -1.0, false);
                    if (player == null) {
                        return;
                    }
                    final ApothBoss item = (ApothBoss)BossRegistry.INSTANCE.getRandomItem(rand, player.m_36336_(), new Predicate[] { WeightedDynamicRegistry.IDimensional.matches((Level)sLevel.m_6018_()), GameStagesCompat.IStaged.matches(player) });
                    if (item == null) {
                        AdventureModule.LOGGER.error("Attempted to spawn a boss in dimension {} using configured boss spawn rule {}/{} but no bosses were made available.", (Object)dimId, (Object)rules.getRight(), (Object)rules.getLeft());
                        return;
                    }
                    final Mob boss = item.createBoss(sLevel, BlockPos.m_274561_(e.getX() - 0.5, e.getY(), e.getZ() - 0.5), rand, player.m_36336_());
                    if (AdventureConfig.bossAutoAggro && !player.m_7500_()) {
                        boss.m_6710_((LivingEntity)player);
                    }
                    if (canSpawn((LevelAccessor)sLevel, boss, player.m_20280_((Entity)boss))) {
                        sLevel.m_47205_((Entity)boss);
                        e.setResult(Event.Result.DENY);
                        AdventureModule.debugLog(boss.m_20183_(), "Surface Boss - " + boss.m_7755_().getString());
                        final Component name = this.getName(boss);
                        if (name == null || name.m_7383_().m_131135_() == null) {
                            AdventureModule.LOGGER.warn("A Boss {} ({}) has spawned without a custom name!", (Object)boss.m_7755_().getString(), (Object)EntityType.m_20613_(boss.m_6095_()));
                        }
                        else {
                            sLevel.m_6907_().forEach(p -> {
                                new Vec3(boss.m_20185_(), AdventureConfig.bossAnnounceIgnoreY ? p.m_20186_() : boss.m_20186_(), boss.m_20189_());
                                final Vec3 vec3;
                                final Vec3 tPos = vec3;
                                if (p.m_20238_(tPos) <= AdventureConfig.bossAnnounceRange * AdventureConfig.bossAnnounceRange) {
                                    final ServerGamePacketListenerImpl f_8906_ = ((ServerPlayer)p).f_8906_;
                                    new ClientboundSetActionBarTextPacket((Component)Component.m_237110_("info.apotheosis.boss_spawn", new Object[] { name, (int)boss.m_20185_(), (int)boss.m_20186_() }));
                                    final ClientboundSetActionBarTextPacket clientboundSetActionBarTextPacket;
                                    f_8906_.m_9829_((Packet)clientboundSetActionBarTextPacket);
                                    final TextColor color = name.m_7383_().m_131135_();
                                    final SimpleChannel channel = Apotheosis.CHANNEL;
                                    new BossSpawnMessage(boss.m_20183_(), (color == null) ? 16777215 : color.m_131265_());
                                    final BossSpawnMessage bossSpawnMessage;
                                    PacketDistro.sendTo(channel, (Object)bossSpawnMessage, player);
                                }
                                return;
                            });
                        }
                        this.bossCooldowns.put((Object)entity.m_9236_().m_46472_().m_135782_(), AdventureConfig.bossSpawnCooldown);
                    }
                }
            }
        }
    }
    
    @Nullable
    private Component getName(final Mob boss) {
        return boss.m_20199_().filter(e -> e.getPersistentData().m_128441_("apoth.boss")).findFirst().map((Function<? super Object, ? extends Component>)Entity::m_7770_).orElse(null);
    }
    
    @SubscribeEvent(priority = EventPriority.LOW)
    public void minibosses(final MobSpawnEvent.FinalizeSpawn e) {
        final LivingEntity entity = (LivingEntity)e.getEntity();
        final RandomSource rand = e.getLevel().m_213780_();
        if (!e.getLevel().m_5776_() && entity instanceof Mob) {
            final Mob mob = (Mob)entity;
            if (e.getResult() != Event.Result.DENY) {
                final ServerLevelAccessor sLevel = e.getLevel();
                final Player player = sLevel.m_45924_(e.getX(), e.getY(), e.getZ(), -1.0, false);
                if (player == null) {
                    return;
                }
                final ApothMiniboss item = (ApothMiniboss)MinibossRegistry.INSTANCE.getRandomItem(rand, player.m_36336_(), new Predicate[] { WeightedDynamicRegistry.IDimensional.matches((Level)sLevel.m_6018_()), GameStagesCompat.IStaged.matches(player), MinibossRegistry.IEntityMatch.matches((Entity)entity) });
                if (item != null && !item.isExcluded(mob, sLevel, e.getSpawnType()) && sLevel.m_213780_().m_188501_() <= item.getChance()) {
                    mob.getPersistentData().m_128359_("apoth.miniboss", MinibossRegistry.INSTANCE.getKey((CodecProvider)item).toString());
                    mob.getPersistentData().m_128350_("apoth.miniboss.luck", player.m_36336_());
                    if (!item.shouldFinalize()) {
                        e.setCanceled(true);
                    }
                }
            }
        }
    }
    
    @SubscribeEvent(priority = EventPriority.LOWEST)
    public void delayedMinibosses(final EntityJoinLevelEvent e) {
        if (!e.getLevel().f_46443_) {
            final Entity entity = e.getEntity();
            if (entity instanceof final Mob mob) {
                final String key = mob.getPersistentData().m_128461_("apoth.miniboss");
                if (key != null) {
                    final ApothMiniboss item = (ApothMiniboss)MinibossRegistry.INSTANCE.getValue(new ResourceLocation(key));
                    if (item != null) {
                        item.transformMiniboss((ServerLevelAccessor)e.getLevel(), mob, e.getLevel().m_213780_(), mob.getPersistentData().m_128457_("apoth.miniboss.luck"));
                    }
                }
            }
        }
    }
    
    @SubscribeEvent
    public void tick(final TickEvent.LevelTickEvent e) {
        if (e.phase == TickEvent.Phase.END) {
            this.bossCooldowns.computeIntIfPresent((Object)e.level.m_46472_().m_135782_(), (key, value) -> Math.max(0, value - 1));
        }
    }
    
    @SubscribeEvent
    public void load(final ServerStartedEvent e) {
        e.getServer().m_129880_(Level.f_46428_).m_8895_().m_164861_((Function)this::loadTimes, () -> new TimerPersistData(), "apotheosis_boss_times");
    }
    
    private TimerPersistData loadTimes(final CompoundTag tag) {
        this.bossCooldowns.clear();
        for (final String s : tag.m_128431_()) {
            final ResourceLocation id = new ResourceLocation(s);
            final int val = tag.m_128451_(s);
            this.bossCooldowns.put((Object)id, val);
        }
        return new TimerPersistData();
    }
    
    private static boolean canSpawn(final LevelAccessor world, final Mob entity, final double playerDist) {
        return (playerDist <= entity.m_6095_().m_20674_().m_21611_() * entity.m_6095_().m_20674_().m_21611_() || !entity.m_6785_(playerDist)) && entity.m_5545_(world, MobSpawnType.NATURAL) && entity.m_6914_((LevelReader)world);
    }
    
    private class TimerPersistData extends SavedData
    {
        public CompoundTag m_7176_(final CompoundTag tag) {
            for (final Object2IntMap.Entry<ResourceLocation> e : BossEvents.this.bossCooldowns.object2IntEntrySet()) {
                tag.m_128405_(((ResourceLocation)e.getKey()).toString(), e.getIntValue());
            }
            return tag;
        }
    }
    
    public enum BossSpawnRules implements BiPredicate<ServerLevelAccessor, BlockPos>
    {
        NEEDS_SKY(BlockAndTintGetter::m_45527_), 
        NEEDS_SURFACE((level, pos) -> pos.m_123342_() >= level.m_6924_(Heightmap.Types.MOTION_BLOCKING_NO_LEAVES, pos.m_123341_(), pos.m_123343_())), 
        BELOW_SURFACE((level, pos) -> pos.m_123342_() < level.m_6924_(Heightmap.Types.MOTION_BLOCKING_NO_LEAVES, pos.m_123341_(), pos.m_123343_())), 
        CANNOT_SEE_SKY((level, pos) -> !level.m_45527_(pos)), 
        SURFACE_OUTER_END((level, pos) -> BossSpawnRules.NEEDS_SURFACE.test(level, pos) && (Mth.m_14040_(pos.m_123341_()) > 1024 || Mth.m_14040_(pos.m_123343_()) > 1024)), 
        ANY((level, pos) -> true);
        
        public static final Codec<BossSpawnRules> CODEC;
        BiPredicate<ServerLevelAccessor, BlockPos> pred;
        
        private BossSpawnRules(final BiPredicate<ServerLevelAccessor, BlockPos> pred) {
            this.pred = pred;
        }
        
        @Override
        public boolean test(final ServerLevelAccessor t, final BlockPos u) {
            return this.pred.test(t, u);
        }
        
        static {
            CODEC = PlaceboCodecs.enumCodec((Class)BossSpawnRules.class);
        }
    }
}
